/* Copyright (c) 1991 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material 
for any purpose and without fee is hereby granted, provided 
that the above copyright notice and this permission notice 
appear in all copies, and that the name of Bellcore not be 
used in advertising or publicity pertaining to this 
material without the specific, prior written permission 
of an authorized representative of Bellcore.  BELLCORE 
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
*/

/* environment.c - contains (nearly) everything for the environmenthandling
   
   Author: Michael Moerz <mikem@debian.org>
*/

#include "metamail.h"
#include "environment.h"

/** adds a string to the current environmet variable PATH
    @param auxpath path to add to $PATH (=auxiliary path)
*/
void
addToPath(auxpath)
char *auxpath;
{
  if (auxpath && *auxpath) {    /* when pointer and auxpath[0] are != NULL */
    static char *newpath = 0;    /* is needed! since it becomes part ov ENV */
    char *oldpath = newpath;
    char *path = getenv("PATH");
    
    if (!path) path = "";	 /* give a default if no current path */
    newpath = XMALLOC(char, (7 + strlen(path) + strlen(auxpath)));
    sprintf(newpath, "PATH=%s:%s", auxpath, path);
    putenv(newpath);
    if (oldpath) XFREE(oldpath); /* free up any old allocated PATH */
  }
}

/** initialize global variable mailheaders from environment
    this code is needed at several places -> own function
*/
void
initMailHeadersFromEnvironment() {
  mailheaders = getenv("MM_HEADERS");
  if (mailheaders) {
    char *s;
    s = XMALLOC(char, 15+strlen(mailheaders));
    sprintf(s, "MM_HEADERS=%s", mailheaders);
    /* error? free mailheaders? */
    mailheaders = s;
  }
}  

/** initializes variables from environment variables or sets them to default
    values if they are not found in the environment
*/
void
initVarsFromEnvironment() {
  char *QuietStr = NULL, *NoAskStr = NULL;

  /* temporary directory - where all temporary files are created */
  tmproot = getenv("METAMAIL_TMPDIR");
  if (!tmproot) {
    tmproot = XMALLOC(char, 5);
    strcpy(tmproot, "/tmp");
  }

  /* */
  initMailHeadersFromEnvironment();

  /* the quietstring that can turn off output for various things */
  QuietStr = getenv("MM_QUIET");
  if (!QuietStr) /* when not set in the environment set default */
    QuietStr = QuietDefault;
  /* now either we are always quiet (="1") or we have to create a structure
     containing information about when we have to be quiet
  */
  if (!strcmp(QuietStr, "1")) { /* when Quietstr == "1" */
    Quiet = 1;
  } else {
    struct NoAskItem *qitem;
    char *s, *tmp;
    char *QuietCopy;

    Quiet = 0;
    QuietCopy = XMALLOC(char, 1+strlen(QuietStr));
    strcpy(QuietCopy, QuietStr);
    for (tmp=QuietCopy; *tmp; ++tmp) {
      if (isupper((unsigned char) *tmp)) 
	*tmp = tolower((unsigned char) *tmp);
    }
    do {
      s = index(QuietCopy, ',');
      if (s) *s++ = 0;
      qitem = XMALLOC(struct NoAskItem, 1);
      qitem->next = FirstQuietItem;
      qitem->type = QuietCopy;
      FirstQuietItem = qitem;
      QuietCopy = s;
    } while (QuietCopy);
  }

  /* */
  if (getenv("MM_TRANSPARENT")) {
    /* Will not propogate recursively */
    TransparentMode = atoi(getenv("MM_TRANSPARENT"));
  }

  /* override for allowing to run as root -- DANGER WILL ROBINSON */
  if (getenv("MM_RUNASROOT")) {
    RunAsRootOK = atoi(getenv("MM_RUNASROOT"));
  }

  /* */
  if (getenv("MM_YANKMODE")) {
    /* Will not propogate recursively */
    YankMode = atoi(getenv("MM_YANKMODE"));
  }

  /* debug flag */
  if (getenv("MM_DEBUG")) {
    DoDebug = atoi(getenv("MM_DEBUG"));
  }

  /* put that somewhere else */
  if (DoDebug) 
    printf("Metamail Version %s, debugging turned on.\n", mmversion);

  /* similar to quiet */
  NoAskStr = getenv("MM_NOASK");
  if (!NoAskStr) /* when noask not set in the ENV then init to default */
    NoAskStr = NoAskDefault;
  if (!strcmp(NoAskStr, "1")) {
    MightAskBeforeExecuting = 0;
  } else {
    struct NoAskItem *nai;
    char *s, *tmp;
    char *NoAskCopy;
    
    NoAskCopy = XMALLOC(char, 1+strlen(NoAskStr));
    strcpy(NoAskCopy, NoAskStr);
    for (tmp=NoAskCopy; *tmp; ++tmp) {
      if (isupper((unsigned char) *tmp)) *tmp = tolower((unsigned char) *tmp);
    }
    do {
      s = index(NoAskCopy, ',');
      if (s) *s++ = 0;
      nai = XMALLOC(struct NoAskItem, 1);
      nai->next = FirstNoAskItem;
      nai->type = NoAskCopy;
      FirstNoAskItem = nai;
      NoAskCopy = s;
    } while (NoAskCopy);
  }

  /* hmm, perhaps mailprogram name ? */
  MailerName = getenv("MM_MAILER");
  if (!MailerName)
    MailerName = "unknown";

  /* */
  if (getenv("MM_USEPAGER")) {
    MaybePageOutput = atoi(getenv("MM_USEPAGER"));
  }

  /* setting if we have a tty or not */
  if ((getenv("MM_NOTTTY") && ((atoi(getenv("MM_NOTTTY"))) != 0))) {
    MustNotBeTty = 1;
  }
  if (MustNotBeTty
      || !isatty(0)
      || !isatty(1)) {
    DefinitelyNotTty = 1;
  }
  
  
}
